-- ESE (Event System for Euphoria) version 1.2
-- by Jeffrey Fielding (JJProg@cyberbury.net)
-- New in 1.2:
-- * Multi-tasking handlers
-- New in 1.1:
-- * Enable/disable/remove sequences of handlers (EnableEventHandlers(handlers))
-- * Push/pop handlers for an event (PushHandlers,PopHandlers)
-- * Remove all handlers for an event (RemoveAllHandlers)
-- * Disable/enable all handlers for an event (Enable/DisableAllHandlers)
-- * Event generators can be passed arguments using
--   AddAdvancedGenerator(rotuine, arguments)
-- * InitESE() clears everything
sequence generators, handlers, handlerids, handlerstack, handlerqueue
integer go
generators = {}
handlers = {}
handlerids = {}
handlerstack = {}
handlerqueue = {}
global procedure InitESE()
    generators = {}
    handlers = {}
    handlerids = {}
    handlerstack = {}
    handlerqueue = {}
end procedure
global procedure PushHandlers(integer event)
    if event > 0 and event <= length(handlers) then
	if sequence(handlers[event]) then
	    handlerstack[event] = append(handlerstack,handlers[event])
	end if
    end if
end procedure
global procedure PopHandlers(integer event)
    if event > 0 and event <= length(handlers) then
	if sequence(handlers[event]) then
	    if length(handlerstack[event]) then
		handlers[event] = handlerstack[event][length(handlerstack)]
		handlerstack[event] = handlerstack[event][1..length(handlerstack)-1]
	    end if
	end if
    end if
end procedure
function NewHandlerId(integer eventid, integer handlerindex)
    integer i
    i = find(-1,handlerids)
    if i then
	handlerids[i] = {eventid,handlerindex}
	return i
    else
	handlerids = append(handlerids,{eventid,handlerindex})
	return length(handlerids)
    end if
end function
global function AddEvent()
    integer i
    i = find(-1,handlers)
    if i then
	handlers[i] = {}
	handlerstack[i] = {}
	return i
    else
	handlers = append(handlers,{})
	handlerstack = append(handlerstack,{})
	return length(handlers)
    end if
end function
global procedure RemoveEvent(integer id)
    if id >= 0 and id <= length(handlers) then
	handlers[id] = -1
	handlerstack[id] = -1
    end if
end procedure
global function AddEventGenerator(integer routine)
    integer i
    if routine = -1 then
	return -1
    end if
    i = find(-1,generators)
    if i then
	generators[i] = {routine,1,{}}
	return i
    else
	generators = append(generators,{routine,1,{}})
	return length(generators)
    end if
end function
global function AddAdvancedGenerator(integer routine, sequence arguments)
    integer i
    if routine = -1 then
	return -1
    end if
    i = find(-1,generators)
    if i then
	generators[i] = {routine,1,arguments}
	return i
    else
	generators = append(generators,{routine,1,arguments})
	return length(generators)
    end if
end function
global function AddEventHandler(integer eventid, integer routine)
    integer i
    if routine = -1 then
	return -1
    end if
    if eventid < 0 or eventid > length(handlers) then
	return -2
    end if
    if integer(handlers[eventid]) then
	return -2
    end if
    i = find(-1,handlers[eventid])
    if i then
	handlers[eventid][i] = {routine,1,0}
	return NewHandlerId(eventid,i)
    else
	handlers[eventid] = append(handlers[eventid],{routine,1,0})
	return NewHandlerId(eventid,length(handlers[eventid]))
    end if
end function
global function AddMultitaskingEventHandler(integer eventid, integer routine)
    integer i
    if routine = -1 then
	return -1
    end if
    if eventid < 0 or eventid > length(handlers) then
	return -2
    end if
    if integer(handlers[eventid]) then
	return -2
    end if
    i = find(-1,handlers[eventid])
    if i then
	handlers[eventid][i] = {routine,1,1}
	return NewHandlerId(eventid,i)
    else
	handlers[eventid] = append(handlers[eventid],{routine,1,1})
	return NewHandlerId(eventid,length(handlers[eventid]))
    end if
end function
global procedure RemoveEventGenerator(integer id)
    if id > 0 and id <= length(generators) then
	generators[id] = -1
    end if
end procedure
global procedure RemoveEventHandler(integer id)
    if id > 0 and id <= length(handlerids) then
	if sequence(handlerids[id]) then
	    if sequence(handlers[handlerids[id][1]]) then
		handlers[handlerids[id][1]][handlerids[id][2]] = -1
		handlerids[id] = -1
	    end if
	end if
    end if
end procedure
global procedure RemoveEventHandlers(sequence s)
    for i = 1 to length(s) do
	RemoveEventHandler(s[i])
    end for
end procedure
global procedure RemoveAllHandlers(integer event)
    if event > 0 and event <= length(handlers) then
	if sequence(handlers[event]) then
	    for i = 1 to length(handlerids) do
		if sequence(handlerids[i]) then
		    if handlerids[i][1] = event then
			RemoveEventHandler(i)
		    end if
		end if
	    end for
	end if
    end if
end procedure
global procedure EnableEventGenerator(integer id)
    if id > 0 and id <= length(generators) then
	generators[id][2] = 1
    end if
end procedure
global procedure EnableEventHandler(integer id)
    if id > 0 and id <= length(handlerids) then
	if sequence(handlerids[id]) then
	    handlers[handlerids[id][1]][handlerids[id][2]][2] = 1
	end if
    end if
end procedure
global procedure EnableAllHandlers(integer event)
    if event > 0 and event <= length(handlers) then
	if sequence(handlers[event]) then
	    for i = 1 to length(handlerids) do
		if sequence(handlerids[i]) then
		    if handlerids[i][1] = event then
			EnableEventHandler(i)
		    end if
		end if
	    end for
	end if
    end if
end procedure
global procedure EnableEventHandlers(sequence ids)
    for i = 1 to length(ids) do
	EnableEventHandler(ids[i])
    end for
end procedure
global procedure DisableEventGenerator(integer id)
    if id > 0 and id <= length(generators) then
	generators[id][2] = 0
    end if
end procedure
global procedure DisableEventHandler(integer id)
    if id > 0 and id <= length(handlerids) then
	if sequence(handlerids[id]) then
	    handlers[handlerids[id][1]][handlerids[id][2]][2] = 0
	end if
    end if
end procedure
global procedure DisableAllHandlers(integer event)
    if event > 0 and event <= length(handlers) then
	if sequence(handlers[event]) then
	    for i = 1 to length(handlerids) do
		if sequence(handlerids[i]) then
		    if handlerids[i][1] = event then
			DisableEventHandler(i)
		    end if
		end if
	    end for
	end if
    end if
end procedure
global procedure DisableEventHandlers(sequence ids)
    for i = 1 to length(ids) do
	DisableEventHandler(ids[i])
    end for
end procedure
global procedure FireEvent(integer id, sequence arguments)
    sequence toCall
    if id > 0 and id <= length(handlers) then
	if sequence(handlers[id]) then
	    toCall = {}
	    for i = 1 to length(handlers[id]) do
		if sequence(handlers[id][i]) then
		    if handlers[id][i][2] then
			if handlers[id][i][3] then
			    handlerqueue = append(handlerqueue,{handlers[id][i][1],arguments})
			else
			    toCall = append(toCall,handlers[id][i][1])
			end if
			--call_proc(handlers[id][i][1],arguments)
		    end if
		end if
	    end for
	    for i = 1 to length(toCall) do
		call_proc(toCall[i],arguments)
	    end for
	end if
    end if
end procedure
global procedure StartEventLoop()
    integer ncalls
    object r
    go = 1
    while go do
	ncalls = 0
	for i = 1 to length(generators) do
	    if sequence(generators[i]) then
		if generators[i][2] then
		    ncalls += 1
		    call_proc(generators[i][1],generators[i][3])
		end if
	    end if
	end for
	if length(handlerqueue) then
	    ncalls += 1
	    r = call_func(handlerqueue[1][1],handlerqueue[1][2])
	    handlerqueue = handlerqueue[2..length(handlerqueue)]
	    if sequence(r) then
		handlerqueue = append(handlerqueue,r)
	    end if
	end if
	if ncalls = 0 then
	    exit
	end if
    end while
end procedure
global procedure StopEventLoop()
    go = 0
end procedure
